/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
/*
 * Miscellaneous functions
 */

#include <assert.h>
#include <limits.h>
#include <ncurses.h>
#include <sys/types.h>
#include <regex.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

#include "main.h"

#include "curs_lib.h"
#include "keymap.h"
#include "lib.h"
#include "misc.h"
#include "options.h"
#include "playlist.h"

/* path of our current working directory */
static char cwd_var[PATH_MAX + 1];

static void change_cwd_var(const char *path);

/*
 * Copy the value of `path', which must be the absolute path to a directory,
 * into `cwd_var', which will then be stripped for trailing slashes if not the
 * root directory.
 */
static void change_cwd_var(const char *path)
{
    assert(path && *path == '/');
    strfcpy(cwd_var, path, sizeof(cwd_var));
    STRIP_TRAILING_SLASHES_IF_NOT_ROOT_DIR(cwd_var);
 }

/*
 * Return a pointer to a static buffer containing the absolute path of the
 * current working directory, which, if not the root directory, will never
 * end with slash.
 */
const char *cwd(void)
{
    return cwd_var;
}

/*
 * Display error message describing the error code passed in the `errcode'
 * parameter, which must be a return value from regcomp() or regexec().
 */
void display_regexp_error(int errcode, const regex_t *preg)
{
     char *errbuf;
     size_t errbuf_size;

     /* Get size of buffer needed for error message */
     errbuf_size = regerror(errcode, preg, NULL, 0);
     /* Allocate buffer and fill it in with error message */
     errbuf = safe_malloc(errbuf_size);
     regerror(errcode, preg, errbuf, errbuf_size);
     msg(errbuf);
     free(errbuf);
}

/*
 * Ask the user for a directory to change to, then xchdir() to it.
 * If an error occurs, tell the user about it.
 * Change the value of the `redraw' member of `menu' as needed.
 * This function is just about the same as browser_chdir().
 */
void main_chdir(struct menu *menu)
{
    char path[PATH_MAX + 1];
    struct stat st;

    snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");

    if (enter_path(path, sizeof(path), DIRECTORY | MUST_EXIST | HALF_FILLED,
		   "Change directory to: ") == NULL)
	return;

    expand_path(path, sizeof(path));

      /* We don't bother checking for dangling symlinks and symlinks to
	 non-directories. */
    if (stat(path, &st) == 0) {
	if (S_ISDIR(st.st_mode)) {
	    if (xchdir(path) != 0) {
		perror_msg("Can't change directory to %s", path);
		return;
	    }
	    TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
	} else
	    msg("%s is not a directory.", path);
    } else
	perror_msg("Can't stat %s", path);
}

void main_make_menu_entry(char *str, size_t len, int index)
{
    int i = 0;
    struct playlist_entry *p = first_entry;

    if (index != 0)
	while (i++ < index)
	    p = p->next;

    snprintf(str, len, "%4d %s %s", index + 1,
	     p->toggled ? "[*]" : "[ ]",
	     is_boolean_option_true(OPT_SHOW_PATH) ? p->path : p->song);
}

/*
 * Return TRUE if `fname' ends with one of the filename extensions defined in
 * the `OPT_FILENAME_EXTENSIONS' option, else return FALSE.
 * The filename extensions are case insensitive.
 */
bool has_registered_filename_extension(const char *fname)
{
    char *p, *q, *r;
    char buf[LONG_STRING], s[LONG_STRING];

    if (!is_string_option_set(OPT_FILENAME_EXTENSIONS))
	return FALSE;

    strfcpy(buf, string_option_value(OPT_FILENAME_EXTENSIONS), sizeof(buf));

    /* here is an example of the OPT_FILENAME_EXTENSIONS option
       ".mp3|mp3player:.midi|midiplayer:.wav|wavplayer" */
    p = buf;
    q = strchr(buf, '|');
    if (q)
	*q++ = '\0';
    q = strchr(q, ':');
    if (q)
	q++;

    strfcpy(s, fname, sizeof(s));
    while (p) {			/* p points to a filename extension */
	r = s;
	while ((r = strstr_icase(r, p))) {
	    if (strlen(r) == strlen(s) - (r - s))
		return TRUE;
	}
	if (!q || !*q) /* if there are no more elements of type .ext|player, */
	  return FALSE;
	p = q;
	q = strchr(q, '|');
	if (q)
	    *q++ = '\0';
	q = strchr(q, ':');
	if (q)
	    q++;
    }

    /* not reached */
    return FALSE;
}

/*
 * If the end of `filename' is found in the `OPT_FILENAME_EXTENSIONS' option,
 * copy the sound player bound to it into `player' and return a pointer to it,
 * else return NULL.
 */
char *extract_player_for_file(const char *filename, char *player,
				size_t playerlen)
{
    char *p, *q, *r, *t;
    char buf[LONG_STRING];
    char fext[LONG_STRING];

    if (!is_string_option_set(OPT_FILENAME_EXTENSIONS))
	return FALSE;

    strfcpy(buf, string_option_value(OPT_FILENAME_EXTENSIONS), sizeof(buf));

    /* here is an example of the OPT_FILENAME_EXTENSIONS option
       ".mp3|mp3player:.midi|midiplayer:.wav|wavplayer" */
    p = buf;
    q = strchr(buf, '|');
    if (!q || !*(q + 1))
	return NULL;
    *q++ = '\0';
    t = q;
    q = strchr(q, ':');
    if (q)
	*q++ = '\0';

    strfcpy(fext, filename, sizeof(fext));
    while (p) {			/* p points to the filename extension */
	r = fext;
	while ((r = strstr_icase(r, p))) {
	    if (strlen(r) == strlen(fext) - (r - fext)) {
		strfcpy(player, t, playerlen);
		return player;
	    }
	}
	if (!q || !*q) /* if there are no more elements of type .ext|player */
	  return NULL;
	/* except of q swapped with buf in the two next statements
	   the next lines are the same as the lines above the loop */
	p = q;
	q = strchr(q, '|');
	if (!q || !*(q + 1))
	    return NULL;
	*q++ = '\0';
	t = q;
	q = strchr(q, ':');
	if (q)
	    q++;
    }

    return NULL;
}

/*
 * The function's name should be self-explaining.
 */ 
bool has_the_correct_syntax_of_the_filename_extensions_variable(char *s)
{
    char *p;

    if (!s || !*s)
	return FALSE;

    p = s;
    while (1) {
	p = strchr(p, '|');
	if (!p || !*(p + 1))	/* if '|' was not found, or '|' was the last character */
	    break;
	p = strchr(p, ':');
	if (!p)
	    return TRUE;	/* success, syntax was perfect */
	p++;			/* make p point to the first character after ':' */
	if (!*p)		/* if colon is the last character */
	    break;
    }
    /* reached if we break the loop */

    return FALSE;
}

/*
 * Return number of help entries for `menu'.
 */
int number_of_menu_help_entries(int menu)
{
    int i = 0, count = 0;

    while (keymap[i].cmd_id != KEYMAP_TERMINATOR) {
	if (keymap[i].menus & menu)
	    count++;
	i++;
    }

    return count;
}

/*
 * Exit the program. Give the user a chance to return if the
 * `CONFIRM_QUIT_CMD' option is TRUE. If `playlist_changed' is TRUE, give
 * the user a chance to save the playlist.
 */
void quit(struct menu *menu, char *playlist, bool playlist_changed)
{
    if (is_boolean_option_true(OPT_CONFIRM_QUIT_CMD))
	if (yes_or_no(1, "Quit?") == FALSE)
	    return;

    if (playlist_changed == TRUE) {
	if (yes_or_no(1, "Save current playlist?")) {
	    int ret;
	    char path[PATH_MAX + 1];
	again:
	    ret = -1;
	    if (playlist) {	/* if we already have a playlist open */
		ret = save_playlist(playlist, OVERWRITE);
		if (ret == 0)	/* if success */
		    msg("Wrote playlist to %s.", playlist);
		else		/* the user will be prompted with "Save playlist to: " the next time */
		    safe_free(playlist);
	    } else if (enter_path(path, sizeof(path), REG_FILE, "Save playlist to: ")) {
		ret = save_playlist(path, ASK_IF_EXISTS);
		if (ret == 0)
		    msg("Wrote playlist to %s.", path);
	    } else		/* user aborted enter_path() */
		ret = -1;

	    if (ret == -1) {
		CLEARLINE(MSG_LINE);
		refresh();
		goto again;
	    }
	    if (ret == 0) {	/* let the user see the "wrote ..." */
		refresh();	/* this refresh is needed */
		putchar('\n');
	    }
	}
    }
    CLEARLINE(MSG_LINE);
    refresh();
    endwin();
    exit(EXIT_SUCCESS);
}

/*
 * After this function, the cursor will be placed at the end of `MSG_LINE',
 * since there's no way to find out its original position.
 */
void sighandler(int sig)
{
    int i;
    chtype ch[STRING];
    
    /* Save the content of the message line */
    for (i = 0; i < COLS && i < sizeof(ch); i++)
	ch[i] = mvinch(MSG_LINE, i);

    if (yes_or_no(1, "Exit [changes will not be saved]?")) {
	CLEARLINE(MSG_LINE);
	refresh();
	endwin();
	exit(EXIT_SUCCESS);
    } else {
	if (signal(SIGINT, sighandler) == SIG_ERR)
	    msg("Warning: re-establishing of the signal handler failed.");
    /* Write the old content of the message line back */
	for (i = 0; i < COLS && i < sizeof(ch); i++)
	    mvaddch(MSG_LINE, i, ch[i]);
	refresh();
    }
}

/*
 * The same as chdir(2) except that this function calls change_cwd_var()
 * if chdir succeeds.
 */
int xchdir(const char *path)
{
  int ret;
  
  ret = chdir(path);
  if (ret == 0)
    change_cwd_var(path);

  return ret;
}

